import json
import logging
from time import sleep, time
from datetime import datetime

from .comm_package import validate_data, define_package, data_to_parsed_string
from .serialib import SerialComm, autodetermine_port
from .const import *
from .exceptions import NoAnswer, InvalidAnswer, InvalidCommand, CRCError

logger = logging.getLogger(__name__)


class Ziva(SerialComm):
    def __init__(self):
        super().__init__()
        self.port = autodetermine_port()
        self.baudrate = 38400
        self.timeout = 0.1
        self.package_counter = 0
        self.rv_units = []
        self.rv_names = []
        self.real_values = []
        self.info_params = {}
        # keeps info and real values together in same var
        self.data = {'real_values':self.real_values, 'info_params':self.info_params}

    def set_port(self, port: str, baudrate: int = 38400, timeout: float = 0.5):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.open()

    def __increment_counter(self):
        if self.package_counter == 255:
            self.package_counter = 0
        else:
            self.package_counter += 1

    def set_device(self, recv_address: int, rf: bool = True):
        self.recv_address = recv_address
        self.rf = rf

    def define_package(self, app_cmd:str = None, var_name:str = None, var_val:str = None):
        package = define_package(
            recv_addr=self.recv_address,
            app_cmd=app_cmd,
            var_name=var_name,
            var_val=var_val,
            counter=self.package_counter,
            rf=self.rf
        )
        return package

    def send_receive(self, app_cmd: str, var_name: str = None, var_val: str = None, retry_limit: int = None,
                     timeout:float = 1, stream_channel: bool = False, parse: bool = True) -> dict:
        '''Sent package and receive answer.'''

        if self.ser != None:
            self.ser.timeout = timeout

        for error_count in range(20):
            try:
                self.__increment_counter()
                package = self.define_package(
                    app_cmd=app_cmd,
                    var_name=var_name,
                    var_val=var_val,
                )
                self.write(data=package)
                data = self.read(last_char=END_MARKER_BYTES)
                data = validate_data(data, stream_channel=stream_channel)
                if stream_channel:
                    # data['data'] = data_decompress(data=data['data'])
                    data['data'] = ''.join(chr(i) for i in data['data'])
                else:
                    data['data'] = data_to_parsed_string(data=data['data'])
            except (NoAnswer, InvalidAnswer, CRCError, InvalidCommand, Exception) as e:
                if retry_limit:
                    if retry_limit == error_count:
                        raise
                else:
                    if type(e).__name__ == NoAnswer.__name__:
                        retry = RETRY_NO_ANSWER
                    elif type(e).__name__ == InvalidAnswer.__name__:
                        retry = RETRY_INVALID_ANSWER
                    elif type(e).__name__ == CRCError.__name__:
                        retry = RETRY_CRC_ERROR
                    elif type(e).__name__ == InvalidCommand.__name__:
                        retry = RETRY_INVALID_COMMAND
                    else:
                        retry = 2
                    if error_count == retry:
                        raise
            else:
                return data

    def read_variable(self, name) -> dict:
        return self.send_receive(app_cmd=VALUE_READ, var_name=name)

    def write_variable(self, name, value) -> dict:
        """Write variable.
        If variable is on of [rv_names, rv_units or rv_mask] we need to empty
        real values that were read before
        """
        if name in [RV_MASK, RV_UNITS, RV_NAMES]:
            self.real_values = []
        return self.send_receive(app_cmd=VALUE_WRITE, var_name=name, var_val=value)

    def read_real_values(self) -> list:
        """ Read real values from device
        Possible output from real values variable:
        - '7005968.000','0.00','-83.94'
        - '29/08/2020 18:39:47','25.740','3.42'
        """

        # If real values were not previously read, read other params also (units,mask,names)
        if not self.real_values:
            self.rv_units = []
            self.rv_names = []
            rv_mask = self.read_variable(name=RV_MASK)['data']
            rv_units = self.read_variable(name=RV_UNITS)['data'].split(';')
            rv_names = self.read_variable(name=RV_NAMES)['data'].split(';')
            if rv_mask and rv_units and rv_names:
                for i, mask in enumerate(rv_mask):
                    if bool(int(mask)):
                        self.rv_units.append(rv_units[i])
                        self.rv_names.append(rv_names[i])

        # Read real values
        real_values = self.read_variable(name=REAL_VALUES)['data']
        self.real_values = []

        # First real value can be timestamp
        try:
            datetime.strptime(real_values[0], "%d/%m/%Y %H:%M:%S")
            self.real_values.append({'channel':'Timestamp', 'value':real_values[0], 'unit':None})
            real_values.pop(0)
        except ValueError:
            pass

        for i, val in enumerate(real_values):
            meas = {
                'channel': self.rv_names[i],
                'value': val,
                'unit': self.rv_units[i]
            }
            self.real_values.append(meas)
        return self.real_values

    def read_params(self):
        data = []
        error_count = 0
        max_error_count = 5
        for i in range(500):
            var_name = f'PAR[{i}]'
            try:
                params = self.send_receive(app_cmd=VALUE_READ, var_name=var_name)['data']
                if params == 'End':
                    break
                elif len(params) > 0:
                    params_dict = {}
                    for par in params:
                        if '=' in par:
                            params_dict[par.split('=')[0]] = par.split('=')[1]
                        else:
                            params_dict[par] = True
                    value = self.read_variable(name=params_dict['Name'])['data']
                    params_dict['Value'] = value
                    data.append(params_dict)
            except Exception as e:
                logger.warning(f'Cant read param {var_name}, e:{e}')
                error_count += 1
                if error_count == max_error_count:
                    break
                else:
                    continue
        logger.info(f'Settings parameters read, count:{len(data)}')
        return data

    def memory_status(self):
        data = self.send_receive(app_cmd=MEMORY_STATUS)['data']
        size, free, used = int(data[0]), int(data[1]), int(data[2])
        return {'size':size, 'free': free, 'used':used, 'used_percentage':round(used / size * 100, 1)}

    def read_memory_data(self, filepath: str = 'data.jda'):
        """Read memory from device and save data to filepath"""

        self.send_receive(app_cmd=OPEN_MEMORY_DIR)
        data_out = ''
        while True:
            data = self.send_receive(app_cmd=MEMORY_START_READING, stream_channel=True, timeout=2)
            data_out += data['data']
            if data['status'] == 'OT':
                break
        if data_out:
            with open(filepath, 'w', encoding='latin1') as f:
                f.write(data_out)

    def save_params(self, filepath:str = 'settings.txt'):
        """Save params to filepath"""
        data = self.read_params()
        with open(filepath, 'w') as f:
            f.write(json.dumps(data))

    def load_params(self, filepath: str):
        with open(filepath) as f:
            data = json.load(f)

    def read_info_variables(self):
        """Get info params"""
        data = {
            'ident_a':IDENT_A,
            'ident_b':IDENT_B,
            'ident_c':IDENT_C,
            'id_0': ID_0,
        }

        for key, val in data.items():
            try:
                var = self.read_variable(name=val)['data']
                data[key] = var
            except Exception as e:
                continue
        self.info_params = data

        try:
            data['real_values'] = self.read_real_values()
        except Exception as e:
            data['real_values'] = None

        return data

    def reset_rf_device(self):
        pass

    def reset_cpu(self):
        return self.send_receive(app_cmd=RESET_CPU)

    def goto_sleep(self):
        return self.send_receive(app_cmd=DEVICE_SLEEP)

    def start_wake_up_routine(self):
        start = time()
        app_cmd = f"{WAKE_UP_DEVICE}{self.recv_address},{WAKE_UP_TIME_STEP}"
        try:
            self.send_receive(app_cmd=app_cmd, retry_limit=3, timeout=0.1)
            sleep(WAKE_UP_TIME_TOTAL)
        except NoAnswer:
            raise NoAnswer('USB router is busy or not working properly')
        wake_time = round(time() - start, 1)
        logger.info(f'Device woke up in {wake_time} seconds')

    def go_online(self) -> dict:
        """
        Go online. If remote device is RF start wake_up procedure.
        :return: dict of info parameters
        """
        if self.rf:
            self.start_wake_up_routine()

        info_vars = self.read_info_variables()
        return info_vars




import sys
sys.path.append('..')
import argparse
import logging

from ziva.serialib import autodetermine_port, serial_ports
from ziva.ziva import Ziva
from ziva.const import VALUE_WRITE, VALUE_READ

parser = argparse.ArgumentParser()

parser.add_argument('communication_type', choices=["rf", "ism"], help="Communication type")
parser.add_argument('-addr', type=int, help="Device address we wish to communicate with")
parser.add_argument('-port', help="Serial port. Port is automatically set by default")
parser.add_argument('-baud', default=38400, help="Serial port baudrate")
parser.add_argument('-online', default=True, help="Online communication")
parser.add_argument('-loglevel', default='debug', choices=["debug", "info", "warning", "error", "exception"], help="Loglevel")
parser.add_argument('-terminal_mode', action='store_true', help='Go straight to terminal mode')
parser.add_argument("--list_ports", action="store_true", help="List COM ports")

args = parser.parse_args()
loglevel = args.loglevel.upper()

port = args.port
baud = args.baud
online = args.online
addr = args.addr
comm_type = args.communication_type
rf = True if args.communication_type == 'rf' else False
terminal_mode = args.terminal_mode
list_ports = args.list_ports

logger = logging.getLogger(__name__)

class ZivaCli(Ziva):
    def __init__(self):
        super().__init__()

    def terminal_mode(self):
        """Offline terminal mode"""

        app_cmd = input("Application command:").upper() or None
        var_name = input("Variable name:").upper() or None
        var_val = input("Variable value:") or None
        return self.send_receive(app_cmd=app_cmd, var_name=var_name, var_val=var_val)

    def terminal_mode_write_variable(self):
        """Online write terminal mode"""

        var_name = input("Variable name:").upper() or None
        var_val = input("Variable value:") or None
        if not (var_name and var_val):
            raise Exception('Both variable name and variable value must be defined')
        return self.send_receive(app_cmd=VALUE_WRITE, var_name=var_name, var_val=var_val)

    def terminal_mode_read_variable(self):
        """Online write terminal mode"""

        var_name = input("Variable name:").upper() or None
        if not var_name:
            raise Exception('Variable name is not defined')
        return self.send_receive(app_cmd=VALUE_READ, var_name=var_name)

def get_input_text():
    input_text = 'Press key for specific task: \n' \
                 'o) Go online \n' \
                 'r) Real values \n' \
                 's) Read settings \n' \
                 'i) Info variables \n' \
                 'w) Write variable \n' \
                 'f) Read variable \n' \
                 't) Terminal mode \n' \
                 'p) Go to sleep (RF) \n' \
                 'm) Memory status \n' \
                 'rm) Read memory \n' \
                 'c) Reset CPU \n' \
                 'q) Quit\n'
    return input_text

def main():
    global port, baud, online, addr, comm_type, rf, terminal_mode, list_ports

    if not port:
        if len(serial_ports()) == 0:
            print ('No serial port detected')
            return

        elif len(serial_ports()) > 1:
            print ('Multiple serial ports detected. Please select one:')
            for i, p in enumerate(serial_ports()):
                print (f"{i+1}) {p}")
            port_number = int(input(""))
            port = serial_ports()[port_number - 1].split(' - ')[0]
        else:
            port = autodetermine_port()

    if not addr:
        addr = int(input("Select remote device address:"))

    ziva = ZivaCli()
    ziva.set_port(port=port, baudrate=baud)
    ziva.set_device(recv_address=addr, rf=rf)
    while True:
        try:
            if terminal_mode:
                key = 't'
            else:
                key = input(get_input_text())

            if key == 'o':
                print (ziva.go_online())
            elif key == 'r':
                try:
                    print (ziva.read_real_values())
                except Exception as e:
                    logger.exception(e)
            elif key == 's':
                try:
                    print (ziva.read_params())
                except Exception as e:
                    print (e)
            elif key == 'i':
                print (ziva.read_info_variables())
            elif key == 'p':
                print (ziva.goto_sleep())
            elif key == 'c':
                print (ziva.goto_sleep())
            elif key == 'm':
                print (ziva.memory_status())
            elif key == 'rm':
                print(ziva.read_memory_data())
            elif key == 'w':
                print(ziva.terminal_mode_write_variable())
            elif key == 'f':
                print(ziva.terminal_mode_read_variable())
            elif key == 't':
                while True:
                    try:
                        print (ziva.terminal_mode())
                    except KeyboardInterrupt:
                        if terminal_mode:
                            exit (0)
                        else:
                            break
                    except Exception as e:
                        print (e)
                        # print (traceback.print_exc(file=sys.stdout))
            elif key == 'q':
                break
        except Exception as e:
            print (e)


if __name__ == '__main__':
    formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(format=formatter)
    logging.root.setLevel(loglevel)
    main()



from time import sleep
from serial.tools import list_ports
import serial
import logging


logger = logging.getLogger(__name__)


def comports():
    """Lists available comports"""

    available_ports = []
    ports = list_ports.comports()
    for port in ports:
        available_ports.append(str(port))
    return available_ports


def serial_ports():
    """Filter serial ports from all comports and strip away description of port"""

    ports = comports()
    ser_ports = list(filter(lambda x: 'Serial Port' in x, ports))
    ser_ports = [i.split(' ')[0] for i in ser_ports]
    return ser_ports


def autodetermine_port():
    """Auto select first serial port"""

    ser_ports = serial_ports()
    if ser_ports:
        return serial_ports()[0]


class SerialComm(object):
    def __init__(self):
        self.port = None
        self.baudrate = 38400
        self.timeout = 0.1
        self.n_retry = 3
        self.ser = None

    def open(self):
        """Connect to serial port"""

        if self.port == None:
            raise Exception('You have to set a serial port')

        for i in range(self.n_retry):
            try:
                self.ser = serial.Serial(
                    port=self.port,
                    baudrate=self.baudrate,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                    bytesize=serial.EIGHTBITS,
                    timeout=self.timeout
                )
                self.ser.flushInput()
                logger.debug(f"Serial port opened:{self.port}, baud:{self.baudrate}")
                logger.debug(f"Serial port flushed")
                break
            except Exception as e:
                logger.error(f"Failed to open serial port:{self.port}, error:{e}")
                sleep(1)

    def close(self):
        """Close serial port"""

        self.ser.close()
        self.ser = None

    def reopen_port(self):
        """Reopen serial port. This is sometimes necessary if it stops working."""

        self.open()

    def write(self, data: bytes = None):
        """Write bytes to serial port"""

        if self.ser == None:
            self.open()

        if not data:
            raise ValueError('No bytes to send')

        logger.debug(f"Data sent: {data}")
        if data:
            self.ser.write(data)

    def readline(self):
        line = self.ser.readline()
        return line

    def read(self, last_char=b'\x83'):
        """Read from serial 1 byte at a time. Stop when last_char is received."""

        frame = b""
        while True:
            byte = self.ser.read(1)
            if byte:
                frame += byte
                if byte == last_char:
                    break
            else:
                break
        logger.debug(f"Data received: {frame}")
        return frame



from binascii import unhexlify

from .const import START_MARKER, END_MARKER, SERVER_ADDR, EQUALS_SIGN, WAKE_UP_DEVICE
from .crc import calculate_crc16
from .exceptions import (
    NoAnswer,
    InvalidAnswer,
    CRCError,
    InvalidCommand,
)


def replace_repeating_chars(data: str = None) -> str:
    """Replace possible repeating chars within communication package
    [0x82] -> [0x84][‘2’]
    [0x83] -> [0x84][‘3’]
    [0x84] -> [0x84][‘4’]
    :param data: input communication string
    :return:
    """

    data_out = ''
    if data:
        for i in range(0, len(data), 2):
            ch = data[i:i + 2]
            if i == 0 or i == len(data) - 2:
                data_out += ch
            else:
                if ch == '82':
                    data_out += '84' + ''.join('{:02x}'.format(ord(c)) for c in '2')
                elif ch == '83':
                    data_out += '84' + ''.join('{:02x}'.format(ord(c)) for c in '3')
                elif ch == '84':
                    data_out += '84' + ''.join('{:02x}'.format(ord(c)) for c in '4')
                else:
                    data_out += ch
    return data_out

def data_decompress(data:bytes) -> bytes:
    """Decompress data. Repeating set of chars are compressed into shorter set.
    Set of K equal chars is transformed (K=char, X=number of repeats):
    [F0][X]K ; K=4-239
    [F0][3][K-hi][K-low]X ; K=240-65520
     """

    compress_char = b'\xf0'
    bytes_out = b''
    i = 0
    while i < len(data):
        byte_char = bytes([data[i]])
        if byte_char == compress_char:
            n_repeats = data[i + 1]
            char = chr(data[i + 2])
            char = f"{n_repeats * char}".encode()
            bytes_out = b"".join([bytes_out, char])
            i += 3
        else:
            bytes_out = b"".join([bytes_out, byte_char])
            i += 1

    return bytes_out

def define_package(recv_addr: int, app_cmd: str, var_name: str = None, var_val: str = None, counter: int = 0,
                   rf: bool = False) -> bytes:
    """
    Returns string that represent whole communication package. Output string can be directly sent to device.
    Example: read real values from ism (addr 4) "820004FF80000000008101565252563D89C483"

    :param recv_addr: receiver address (e.g. 1000)
    :param app_cmd: application command (e.g. VR - value read)
    :param var_name: variable name (e.g. RV - real values)
    :param var_val: variable value (e.g. `Merilno mest` or `3600`)
    :param rf: RF or not (wifi or wired connection)
    :param counter: package counter (0-255)
    :return: parsed data as string
    """

    # Generate receiver address from integer recv_addr
    if rf:
        recv_addr_complete = '{:8X}'.format(recv_addr).replace(' ', '0')
    else:
        recv_addr_1 = '{:2X}'.format(0).replace(' ', '0')
        recv_addr_2 = '{:2X}'.format(recv_addr).replace(' ', '0')
        recv_addr_3 = '{:2X}'.format(255).replace(' ', '0')
        recv_addr_4 = '{:2X}'.format(128).replace(' ', '0')
        recv_addr_complete = recv_addr_1 + recv_addr_2 + recv_addr_3 + recv_addr_4

    # Consts
    start_marker = START_MARKER
    end_marker = END_MARKER
    serv_addr = SERVER_ADDR
    equals_sign = EQUALS_SIGN

    # data layer
    data = "".join("{:02x}".format(ord(c)) for c in app_cmd)
    if var_name:
        data += "".join("{:02x}".format(ord(c)) for c in var_name)
        data += equals_sign
    if var_val:
        data += "".join("{:02x}".format(ord(c)) for c in var_val)

    if rf and WAKE_UP_DEVICE in app_cmd:
        data += '{:2X}'.format(0).replace(' ', '0')

    # Get hex counter value and 16bit CRC code
    counter_hex = '81' + '{:2X}'.format(counter).replace(' ', '0')
    data_for_crc = recv_addr_complete + serv_addr + counter_hex + data
    crc16 = calculate_crc16(unhexlify(data_for_crc))

    # Combine all data layers
    package = f"{start_marker}{recv_addr_complete}{serv_addr}{counter_hex}{data}{crc16}{end_marker}"
    package = replace_repeating_chars(data=package)
    return unhexlify(package)


def data_to_parsed_string(data: bytes):
    """Parse data to list. Separate by comma.
    Example of a raw data: b'\x82\x00\x00\x00\x00\x00\x00\x00\x00\x01@O0.638,0.00,21.67,\to\x83
    """
        # print (data)
    data_str = ''.join(chr(i) for i in data)
    if 'Name' and 'Com' in data_str:
        data_str = list(filter(None, data_str.split(',')))
        data_out = []
        for i, item in enumerate(data_str):
            # If first letter after comma is not in the upper case,
            # merge item with previous item
            if item[0].isupper():
                data_out.append(item)
            else:
                data_out[-1] += ',' + item
        data_str = data_out
    else:
        data_str = list(filter(None, data_str.split(',')))

    return data_str[0] if len(data_str) == 1 else data_str


def validate_data(data: bytes, stream_channel: bool = False) -> dict:
    '''Validate data bytes.
    :param data: input bytes
    :param stream_channel: stream channel uses different status_sign (O, E, OT or OM)
    '''

    if not data:
        raise NoAnswer('No answer from device')
    if len(data) < 14:
        raise InvalidAnswer('Data length is not sufficient')

    # data = ''.join(chr(i) for i in data)

    # Remove 82 and 83, server addr, CRC
    start_char = 1
    end_char = 1
    srv_addr = 8
    crc = 2
    data = data[start_char + srv_addr:-(end_char + crc)]

    # Validate data. Possible combinations are:
    # sign @ signals start of message. First letter or first two letters after @ = status:
    # O[value] - okay
    # E[error message] - error
    # OM[data] -> okay, more data
    # OT[data] -> okay, data read complete
    # OE empty dir
    # W[percent] - not yet completed

    at_sign_pos = data.find(b'@')
    if at_sign_pos == -1 or len(data) == at_sign_pos:
        # Error in message if @ is not present or @ is last char in message
        raise InvalidAnswer('Unkown error:', data)

    status_sign_pos = at_sign_pos + 1
    status_sign = chr(data[status_sign_pos])
    if status_sign == 'O' or status_sign == 'W':
        if stream_channel and len(data[status_sign_pos:]) > 1:
            status_sign_additional_letter = chr(data[status_sign_pos + 1])
            if status_sign_additional_letter == 'T' or status_sign_additional_letter == 'M':
                status_sign += status_sign_additional_letter
        data_out = {'status':status_sign, 'data':data[data.find(status_sign.encode()) + len(status_sign):]}
        return data_out

    if status_sign == 'E':
        error_msg = data[status_sign_pos:]
        if b'implemented' in data:
            raise InvalidCommand('Invalid command')

        elif b'Undefined variable' in data:
            raise InvalidCommand('Undefined variable')

        elif b'CRC error' in data:
            raise CRCError('Invalid CRC code')
        else:
            raise InvalidAnswer('Unkown error:', error_msg)
    elif status_sign == ' ':
        if b'implemented' in data:
            raise InvalidCommand('Invalid command')
    else:
        raise InvalidAnswer('Status sign is not valid: ', data)

    # Get status from data
    # try:
    #     status_data = data.split('@')[1]
    #     if len(status_data) > 1:
    #         status = status_data[0] if status_data[1] not in ['M', 'T'] else status_data[0:2]
    #     else:
    #         status = status_data[0]
    # except IndexError:
    #     raise InvalidAnswer('Unkown error:', data)
    # if status in ['O', 'OT', 'OM', 'W']:
    #     message = data.split('@' + status)[1]
    #     if parse:
    #         message = parse_data(data=message)
    #     return {'status': status, 'data': message}
    # elif status == 'E':
    #     if 'implemented' in data:
    #         raise InvalidCommand('Invalid command')
    #
    #     elif 'Undefined variable' in data:
    #         raise InvalidCommand('Undefined variable')
    #
    #     elif 'CRC error' in data:
    #         raise CRCError('Invalid CRC code')
    #
    #     else:
    #         raise InvalidAnswer('Unkown error:', data.split('@E'))
    # else:
    #     raise InvalidAnswer('Unkown error:', data)

    # elif status == 'E':
    #     raise AnswerIsNotValid("Error: " + data.split('@E')[1])
    # elif status == ' ':
    #     raise AnswerIsNotValid("Error: " + data.split('@ ')[1])
    # else:
    #     raise AnswerIsNotValid('Unkown error:', data)



from binascii import hexlify, unhexlify

t = """gÚ´RZ@¶'gÚð\x00RYï¶'gÛ,RZR¶'gÛhRZú¶'gÛ¤RZ~¶'gÛàRZ\x96¶'gÜ\x1cRYü¶'gÜXRZk¶'gÜ\x94RZx¶'gÜÐRZ\x96¶'gÝ\x0cRZÎ¶'gÝHR[1¶'gÝ\x844R[i¶'gÝÀR[\x00¶"""
print (unhexlify(t))


# with open('data.jda', 'w', encoding='latin1') as f:
#     f.write(t)

def jda_to_dtv(filepath) -> str:
    with open(filepath, 'r') as f:
        lines = f.readlines()

    file_out = ''
    return file_out
